import pkg_resources
import pandas as pd

def load_manzana_data():
    stream = pkg_resources.resource_stream(__name__, 'data/reduced_data_MinMaxScaler_PCA_manzana_224.csv')
    return pd.read_csv(stream)


from SpatialCluster.utils.data_format import data_format, position_data_format
from SpatialCluster.utils.get_areas import get_areas
from sklearn.mixture import GaussianMixture

"""
----------------
Parameters:

features_X:         (DataFrame) Table with the features of each point
features_position:  (DataFrame) Table with the position of each point (lon, lat)
n_components:       (int) Number of mixture components

----------------
Return:

areas_to_points:    (dict) Dictionary with the cluster id as keys and a list of points in as values

"""

def GMM_Clustering(features_X, features_position, n_clusters = 2, covariance_type = "full", tol=1e-3, reg_covar=1e-6):
    features_X = data_format(features_X)
    features_position = position_data_format(features_position)
    points = list(zip(features_position.lon, features_position.lat))
    gmm = GaussianMixture(n_components=n_clusters, covariance_type=covariance_type, tol=tol, reg_covar=reg_covar).fit(features_X)
    clusters = gmm.predict(features_X)

    areas_to_points = get_areas(clusters, points)

    return areas_to_points, clusters


from SpatialCluster.methods.DMoN_core import convert_scipy_sparse_to_sparse_tensor, build_dmon, normalize_graph
from SpatialCluster.utils.data_format import data_format, position_data_format
from SpatialCluster.preprocess import adjacencyMatrix
from SpatialCluster.utils.get_areas import get_areas
import tensorflow.compat.v2 as tf
import numpy as np


"""
----------------
Parameters:

features_X:         (DataFrame) Table with the features of each point
features_position:  (DataFrame) Table with the position of each point (lon, lat)
r_max:              (float) Max ratio to search for neighbours
n_clusters:         (int) Number of clusters
reg:                (int) Collapse regularization
dropout:            (float) Dropout rate for the model
num_epochs:         (int) Number of epochs to train the model

----------------
Return:

areas_to_points:    (dict) Dictionary with the cluster id as keys and a list of points in as values

"""

def DMoN_Clustering(features_X, features_position, A = None, criteria = "rk", r_max = 300, n_clusters = 4, reg = 1, dropout = 0.0, num_epochs = 500, learning_rate = 0.001):
    features_X = data_format(features_X)
    features_position = position_data_format(features_position)
    X = features_X.to_numpy(dtype=float)
    x_min, x_max = np.min(X, axis=0), np.max(X, axis=0)
    try:
        X = (X - x_min)/(x_max - x_min)
    except RuntimeWarning:
        X = (X - x_min)/(x_max - x_min)
        print("--------------------Debug------------------------")
        print(f"X:{X}\nx_min:{x_min}\nx_max:{x_max}")
    X = np.asmatrix(X) # feature matrix

    # --------------------------------------------------------------------------
    points = list(zip(features_position.lon, features_position.lat))
    if(A == None):
        A = adjacencyMatrix(features_position, criteria=criteria, r=r_max)
    n_nodes = A.shape[0]
    feature_size = X.shape[1]

    # --------------------------------------------------------------------------

    graph = convert_scipy_sparse_to_sparse_tensor(A)
    graph_normalized = convert_scipy_sparse_to_sparse_tensor(normalize_graph(A.copy()))
    input_features = tf.keras.layers.Input(shape=(feature_size,))
    input_graph = tf.keras.layers.Input((n_nodes,), sparse=True)
    input_adjacency = tf.keras.layers.Input((n_nodes,), sparse=True)

    model = build_dmon(input_features, input_graph, input_adjacency, n_clusters, reg, dropout)

    # Computes the gradients wrt. the sum of losses, returns a list of them.
    def grad(model, inputs):
        with tf.GradientTape() as tape:
            _ = model(inputs, training=True)
            loss_value = sum(model.losses)
        return model.losses, tape.gradient(loss_value, model.trainable_variables)

    optimizer = tf.keras.optimizers.Adam(learning_rate)
    model.compile(optimizer, None)

    # --------------------------------------------------------------------------

    for epoch in range(num_epochs):
        loss_values, grads = grad(model, [X, graph_normalized, graph])
        optimizer.apply_gradients(zip(grads, model.trainable_variables))

    # --------------------------------------------------------------------------

    _, assignments = model([X, graph_normalized, graph], training=False)
    assignments = assignments.numpy()
    clusters = assignments.argmax(axis=1) 

    areas_to_points = get_areas(clusters, points)

    return areas_to_points, clusters



def attributes_format(df):
    position = df[["lon", "lat"]]
    cols = df.columns
    attr_cols = []
    for col in cols:
        if(col != "lon" and col != "lat"):
            if("float" in str(type(df[col][0]))):
                attr_cols.append(col)
    attributes = df[attr_cols]
    return position, attributes

def attributes_with_zone_format(df):
    position = df[["lon", "lat"]]
    cols = df.columns
    attr_cols = []
    for col in cols:
        if(col != "lon" and col != "lat"):    
            if("float" in str(type(df[col][0]))):
                attr_cols.append(col)
            elif(str == type(df[col][0])):
                attr_cols.append(col)
    attributes = df[attr_cols]
    return position, attributes



from SpatialCluster.utils.data_format import numpy_data_format, position_data_format
from SpatialCluster.utils.data_structures  import IncrementalCOOMatrix
from SpatialCluster.preprocess import adjacencyMatrix
from SpatialCluster.utils.get_areas import get_areas
from scipy.spatial.distance import jensenshannon
from sklearn.cluster import spectral_clustering
from sklearn.preprocessing import MinMaxScaler 
import pandas as pd
import numpy as np
import scipy as sp

import time

def TDI_Clustering(features_X, features_position, A = None, r=300, k=5, leafsize=10):
    a = time.time()
    features_X = numpy_data_format(features_X)
    features_position = position_data_format(features_position)
    scaler = MinMaxScaler()
    new_features = pd.DataFrame(scaler.fit_transform(features_X))
    criteria = "k"
    directed = False
    if(A == None):
        A = adjacencyMatrix(features_position, r=r, k=k, criteria=criteria, directed=directed, leafsize=leafsize)
    mat = IncrementalCOOMatrix(A.shape, np.float64)
    rows, cols, values = sp.sparse.find(A)
    b = time.time()
    print("First Part:", b-a)
    c = time.time()
    for i in range(rows.shape[0]):
        v1_pos = rows[i]
        v2_pos = cols[i]
        if(v2_pos > v1_pos):
            continue
        vec1 = np.array(new_features.loc[v1_pos])
        vec2 = np.array(new_features.loc[v2_pos])
        dist = jensenshannon(vec1, vec2)
        if np.isnan(dist):
            dist = 0
        mat.append(v1_pos, v2_pos, dist)
    d = time.time()
    print("Matriz de pesos:", d-c)
    e = time.time()
    mat = mat.tocoo() # weight matrix
    mat = mat.tocsr()
    f = time.time()
    print("Transformar matriz:", f-e)
    g = time.time()
    clusters = spectral_clustering(mat)
    h = time.time()
    print("Spectral clustering:", h-g)
    i = time.time()
    points = list(zip(features_position.lon, features_position.lat))
    areas_to_points = get_areas(clusters, points)
    j = time.time()
    print("Areas to points:", j-i)
    return areas_to_points, clusters


from SpatialCluster.utils.data_structures  import IncrementalCOOMatrix
from SpatialCluster.utils.data_format import position_data_format
import scipy as sp
import numpy as np

def adjacencyMatrix(features_position, r = 300, k = 5, min_k = 2, criteria = "k", directed = True, leafsize = 10):
    meter_to_degree_equivalence = 9.090909091e-6
    r = r*meter_to_degree_equivalence # Se transforma de metros a grados
    features_position = position_data_format(features_position)
    points = list(zip(features_position.lon, features_position.lat))

    # --------------------------------------------------------------------------
    len_points = len(points)
    shape = (len_points, len_points)
    mat = IncrementalCOOMatrix(shape, np.int64)

    tree = sp.spatial.KDTree(data = points, leafsize = leafsize)
    if(criteria == "k"):
        k_neighbours = tree.query(points, k = k)[1] # k vecinos
        for i, neighbourhood in enumerate(k_neighbours):
            for neighbour in neighbourhood:
                mat.append(i, neighbour, 1)
                if(not directed):
                    mat.append(neighbour, i, 1)
    elif(criteria == "r"):
        nearest_neighbours = tree.query_ball_tree(tree, r) # vecinos dentro de radio r
        for i, neighbourhood in enumerate(nearest_neighbours):
            for neighbour in neighbourhood:
                mat.append(i, neighbour, 1)
                if(not directed):
                    mat.append(neighbour, i, 1)
    elif(criteria == "rk"):
        nearest_neighbours = tree.query_ball_tree(tree, r) # vecinos dentro de radio r
        k_neighbours = tree.query(points, k = k)[1] # k vecinos
        for i, neighbourhood in enumerate(nearest_neighbours):
            if len(neighbourhood) < min_k: # si la cantidad de vecinos dentro del radio es menor que min_k
                neighbourhood = k_neighbours[i] # se usan los k vecinos
            for neighbour in neighbourhood:
                mat.append(i, neighbour, 1)
                if(not directed):
                    mat.append(neighbour, i, 1)
    else:
        raise ValueError("Valor inválido para el parámetro 'criteria' ('r', 'k', 'rk')")

    A = mat.tocoo() # adjacency matrix
    rows, cols, values = sp.sparse.find(A)
    for i in range(values.shape[0]):
        values[i] = 1
    A = sp.sparse.coo_matrix((values, (rows, cols)), shape=A.shape)
    A = A.tocoo()
    A = A.tocsr()

    return A



"""
Init file for the SpatialCluster library.
"""

version_info = (0, 0, 69)
__version__ = ".".join([str(x) for x in version_info])

__all__ = [
    "methods"
    "preprocess"
    "visualization"
    "metrics"
    "utils"
]


from SpatialCluster.utils.data_format import numpy_data_format, position_data_format
from SpatialCluster.utils.get_areas import get_areas
from minisom import MiniSom  
import numpy as np

def SOM_Clustering(features_X, features_position, som_shape = (6,6), sigma=0.3, learning_rate=0.5, num_iterations = 100):
    features_X = numpy_data_format(features_X)
    features_position = position_data_format(features_position)
    
    som = MiniSom(som_shape[0], som_shape[1], features_X.shape[1], sigma=sigma, learning_rate=learning_rate) # initialization of SOM
    som.train(features_X, num_iterations)
    winner_coordinates = np.array([som.winner(x) for x in features_X]).T
    # with np.ravel_multi_index we convert the bidimensional
    # coordinates to a monodimensional index
    clusters = np.ravel_multi_index(winner_coordinates, som_shape)
    points = list(zip(features_position.lon, features_position.lat))
    areas_to_points = get_areas(clusters, points)
    return areas_to_points, clusters


from SpatialCluster.utils.data_format import data_format, position_data_format
from scipy import spatial
import pandas as pd
import numpy as np

def rings(features_X, features_position, criteria="k", max_radios=[200, 300, 400], max_neighbours=[200, 500, 1000], weight_mode="Simple", keep_original_value=True, smoothing=1e-08, normalize=True, leafsize=10):

    """
    El algoritmo toma todos los puntos dentro de un radio máximo.
    Luego, toma los K vecinos más cercanos a cada pundo donde K
    está definido según "max_neighbors_per_radio".
    """
    original_radios = max_radios
    meter_to_degree_equivalence = 9.090909091e-6
    max_radios = np.array(max_radios)*meter_to_degree_equivalence # Se transforma de metros a grados
    features_X = data_format(features_X)
    features_position = position_data_format(features_position)
    # Verificar que ambas listas sean de largo igual
    assert len(max_radios) == len(max_neighbours)

    points = np.array(list(zip(features_position.lon, features_position.lat)))
    tree = spatial.KDTree(data=points, leafsize=leafsize)
    final_features_df = pd.DataFrame()

    if keep_original_value:
        final_features_df = features_X.copy()
    len_features = len(features_X.columns)
    if(criteria == "k"):
        for _, k in enumerate(max_neighbours):
            features_array = [ [] for _ in range(len_features)]
            k_neighbours = tree.query(points, k = k)[1] # k vecinos
            for _, neighbourhood in enumerate(k_neighbours):
                for f_index, feature in enumerate(features_X.columns):
                    subset_feature = features_X.iloc[neighbourhood][feature]
                    mean_feature = subset_feature.mean()
                    features_array[f_index].append(mean_feature) 
            for f_index, feature in enumerate(features_X.columns):
                final_features_df[f"{feature}_{k}nb"] = features_array[f_index] # indico cuantos vecinos en el nombre del attributo
        features_X = final_features_df
    elif(criteria == "r"):
        for i, r in enumerate(max_radios):
            features_array = [ [] for _ in range(len_features)]
            nearest_neighbours = tree.query_ball_tree(tree, r) # vecinos dentro de radio r
            for _, neighbourhood in enumerate(nearest_neighbours):
                for f_index, feature in enumerate(features_X.columns):
                    subset_feature = features_X.iloc[neighbourhood][feature]
                    mean_feature = subset_feature.mean()
                    features_array[f_index].append(mean_feature) 
            for f_index, feature in enumerate(features_X.columns):
                final_features_df[f"{feature}_{original_radios[i]}m"] = features_array[f_index]
        features_X = final_features_df
    elif(criteria == "rk"):
        for i, r in enumerate(max_radios):
            features_array = [ [] for _ in range(len_features)]
            k = max_neighbours[i]
            nearest_neighbours = tree.query_ball_tree(tree, r)
            for j, neighbourhood in enumerate(nearest_neighbours): # Para cada punto
                neighbourhood = np.array(neighbourhood) # Saco los vecinos dentro del radio r
                temporal_tree = spatial.KDTree(data=points[neighbourhood], leafsize=leafsize)
                k_length = min(k, len(neighbourhood)) # Conseguir los K puntos más cercanos dentro de ese radio
                distances, nearby_points = temporal_tree.query(points[j], k = k_length)
                distances += smoothing # Tal vez simplemente eliminar el primer elemento que es el punto mismo
                if(normalize):
                    distances = distances/np.linalg.norm(distances)
                indexs = neighbourhood[nearby_points]
                for f_index, feature in enumerate(features_X.columns):
                    if weight_mode == "Simple":
                        subset_feature = features_X.iloc[indexs][feature]
                    elif weight_mode == "Distance Inverse":
                        subset_feature = features_X.iloc[indexs][feature]/distances
                    mean_feature = subset_feature.mean()
                    features_array[f_index].append(mean_feature) 
            
            for f_index, feature in enumerate(features_X.columns):
                final_features_df[f"{feature}_{original_radios[i]}m{k}nb"] = features_array[f_index]
        features_X = final_features_df
    return features_X


def get_areas(clusters, points):
    num_points = len(points)
    areas_to_points = dict()
    for i in range(num_points):
        area = int(clusters[i])
        if area in areas_to_points:
            areas_to_points[area].append(points[i])
        else:
            areas_to_points[area] = [points[i]]

    return areas_to_points


from sklearn.metrics import cluster
import tensorflow.compat.v2 as tf
import scipy as sp
import numpy as np
import array


class DMoN(tf.keras.layers.Layer):
    """Implementation of Deep Modularity Network (DMoN) layer.
    Deep Modularity Network (DMoN) layer implementation as presented in
    "Graph Clustering with Graph Neural Networks" in a form of TF 2.0 Keras layer.
    DMoN optimizes modularity clustering objective in a fully unsupervised mode,
    however, this implementation can also be used as a regularizer in a supervised
    graph neural network. Optionally, it does graph unpooling.
    Attributes:
        n_clusters: Number of clusters in the model.
        collapse_regularization: Collapse regularization weight.
        dropout_rate: Dropout rate. Note that the dropout in applied to the
        intermediate representations before the softmax.
        do_unpooling: Parameter controlling whether to perform unpooling of the
        features with respect to their soft clusters. If true, shape of the input
        is preserved.
    """

    def __init__(self,
                n_clusters,
                collapse_regularization = 0.1,
                dropout_rate = 0,
                do_unpooling = False):
        """Initializes the layer with specified parameters."""
        super(DMoN, self).__init__()
        self.n_clusters = n_clusters
        self.collapse_regularization = collapse_regularization
        self.dropout_rate = dropout_rate
        self.do_unpooling = do_unpooling
        

    def build(self, input_shape):
        """Builds the Keras model according to the input shape."""
        self.transform = tf.keras.models.Sequential([
            tf.keras.layers.Dense(
                self.n_clusters,
                kernel_initializer='orthogonal',
                bias_initializer='zeros'),
            tf.keras.layers.Dropout(self.dropout_rate)
        ])
        super(DMoN, self).build(input_shape)

    def call(
        self, inputs):
        """Performs DMoN clustering according to input features and input graph.
        Args:
        inputs: A tuple of Tensorflow tensors. First element is (n*d) node feature
            matrix and the second one is (n*n) sparse graph adjacency matrix.
        Returns:
        A tuple (features, clusters) with (k*d) cluster representations and
        (n*k) cluster assignment matrix, where k is the number of cluster,
        d is the dimensionality of the input, and n is the number of nodes in the
        input graph. If do_unpooling is True, returns (n*d) node representations
        instead of cluster representations.
        """
        features, adjacency = inputs

        assert isinstance(features, tf.Tensor)
        assert isinstance(adjacency, tf.SparseTensor)
        assert len(features.shape) == 2
        assert len(adjacency.shape) == 2
        assert features.shape[0] == adjacency.shape[0]

        assignments = tf.nn.softmax(self.transform(features), axis=1)
        cluster_sizes = tf.math.reduce_sum(assignments, axis=0)  # Size [k].
        assignments_pooling = assignments / cluster_sizes  # Size [n, k].

        degrees = tf.sparse.reduce_sum(adjacency, axis=0)  # Size [n].
        degrees = tf.reshape(degrees, (-1, 1))

        number_of_nodes = adjacency.shape[1]
        number_of_edges = tf.math.reduce_sum(degrees)

        # Computes the size [k, k] pooled graph as S^T*A*S in two multiplications.
        graph_pooled = tf.transpose(
            tf.sparse.sparse_dense_matmul(adjacency, assignments))
        graph_pooled = tf.matmul(graph_pooled, assignments)

        # We compute the rank-1 normalizer matrix S^T*d*d^T*S efficiently
        # in three matrix multiplications by first processing the left part S^T*d
        # and then multyplying it by the right part d^T*S.
        # Left part is [k, 1] tensor.
        normalizer_left = tf.matmul(assignments, degrees, transpose_a=True)
        # Right part is [1, k] tensor.
        normalizer_right = tf.matmul(degrees, assignments, transpose_a=True)

        # Normalizer is rank-1 correction for degree distribution for degrees of the
        # nodes in the original graph, casted to the pooled graph.
        normalizer = tf.matmul(normalizer_left,
                            normalizer_right) / 2 / number_of_edges
        spectral_loss = -tf.linalg.trace(graph_pooled -
                                        normalizer) / 2 / number_of_edges
        self.add_loss(spectral_loss)

        collapse_loss = tf.norm(cluster_sizes) / number_of_nodes * tf.sqrt(
            float(self.n_clusters)) - 1
        self.add_loss(self.collapse_regularization * collapse_loss)

        features_pooled = tf.matmul(assignments_pooling, features, transpose_a=True)
        features_pooled = tf.nn.selu(features_pooled)
        if self.do_unpooling:
            features_pooled = tf.matmul(assignments_pooling, features_pooled)
        return features_pooled, assignments

# -----------------------------------------------------------------------------

class GCN(tf.keras.layers.Layer):
    """Implementation of Graph Convolutional Network (GCN) layer.
    Attributes:
        n_channels: Output dimensionality of the layer.
        skip_connection: If True, node features are propagated without neighborhood
        aggregation.
        activation: Activation function to use for the final representations.
    """

    def __init__(self,
                n_channels,
                activation='selu',
                skip_connection = True):
        """Initializes the layer with specified parameters."""
        super(GCN, self).__init__()
        self.n_channels = n_channels
        self.skip_connection = skip_connection
        if isinstance(activation, str):
            self.activation = tf.keras.layers.Activation(activation)
        elif isinstance(tf.keras.layers.Activation):
            self.activation = activation
        elif activation is None:
            self.activation = tf.keras.layers.Lambda(lambda x: x)
        else:
            raise ValueError('GCN activation of unknown type')

    def build(self, input_shape):
        """Builds the Keras model according to the input shape."""
        self.n_features = input_shape[0][-1]
        self.kernel = self.add_weight(
            'kernel', shape=(self.n_features, self.n_channels))
        self.bias = self.add_weight('bias', shape=(self.n_channels,))
        if self.skip_connection:
            self.skip_weight = self.add_weight(
                'skip_weight', shape=(self.n_channels,))
        else:
            self.skip_weight = 0
        super().build(input_shape)

    def call(self, inputs):
        """Computes GCN representations according to input features and input graph.
        Args:
        inputs: A tuple of Tensorflow tensors. First element is (n*d) node feature
            matrix and the second is normalized (n*n) sparse graph adjacency matrix.
        Returns:
        An (n*n_channels) node representation matrix.
        """
        features, norm_adjacency = inputs

        assert isinstance(features, tf.Tensor)
        assert isinstance(norm_adjacency, tf.SparseTensor)
        assert len(features.shape) == 2
        assert len(norm_adjacency.shape) == 2
        assert features.shape[0] == norm_adjacency.shape[0]

        output = tf.matmul(features, self.kernel)
        if self.skip_connection:
            output = output * self.skip_weight + tf.sparse.sparse_dense_matmul(
                norm_adjacency, output)
        else:
            output = tf.sparse.sparse_dense_matmul(norm_adjacency, output)
        output = output + self.bias
        return self.activation(output)

# -----------------------------------------------------------------------------



def pairwise_precision(y_true, y_pred):
    """Computes pairwise precision of two clusterings.
    Args:
        y_true: An [n] int ground-truth cluster vector.
        y_pred: An [n] int predicted cluster vector.
    Returns:
        Precision value computed from the true/false positives and negatives.
    """
    true_positives, false_positives, _, _ = _pairwise_confusion(y_true, y_pred)
    return true_positives / (true_positives + false_positives)


def pairwise_recall(y_true, y_pred):
    """Computes pairwise recall of two clusterings.
    Args:
        y_true: An (n,) int ground-truth cluster vector.
        y_pred: An (n,) int predicted cluster vector.
    Returns:
        Recall value computed from the true/false positives and negatives.
    """
    true_positives, _, false_negatives, _ = _pairwise_confusion(y_true, y_pred)
    return true_positives / (true_positives + false_negatives)


def pairwise_accuracy(y_true, y_pred):
    """Computes pairwise accuracy of two clusterings.
    Args:
        y_true: An (n,) int ground-truth cluster vector.
        y_pred: An (n,) int predicted cluster vector.
    Returns:
        Accuracy value computed from the true/false positives and negatives.
    """
    true_pos, false_pos, false_neg, true_neg = _pairwise_confusion(y_true, y_pred)
    return (true_pos + false_pos) / (true_pos + false_pos + false_neg + true_neg)


def _pairwise_confusion(
    y_true,
    y_pred):
    """Computes pairwise confusion matrix of two clusterings.
    Args:
        y_true: An (n,) int ground-truth cluster vector.
        y_pred: An (n,) int predicted cluster vector.
    Returns:
        True positive, false positive, true negative, and false negative values.
    """
    contingency = cluster.contingency_matrix(y_true, y_pred)
    same_class_true = np.max(contingency, 1)
    same_class_pred = np.max(contingency, 0)
    diff_class_true = contingency.sum(axis=1) - same_class_true
    diff_class_pred = contingency.sum(axis=0) - same_class_pred
    total = contingency.sum()

    true_positives = (same_class_true * (same_class_true - 1)).sum()
    false_positives = (diff_class_true * same_class_true * 2).sum()
    false_negatives = (diff_class_pred * same_class_pred * 2).sum()
    true_negatives = total * (
        total - 1) - true_positives - false_positives - false_negatives

    return true_positives, false_positives, false_negatives, true_negatives


def modularity(adjacency, clusters):
    """Computes graph modularity.
    Args:
        adjacency: Input graph in terms of its sparse adjacency matrix.
        clusters: An (n,) int cluster vector.
    Returns:
        The value of graph modularity.
        https://en.wikipedia.org/wiki/Modularity_(networks)
    """
    degrees = adjacency.sum(axis=0).A1
    n_edges = degrees.sum()  # Note that it's actually 2*n_edges.
    result = 0
    for cluster_id in np.unique(clusters):
        cluster_indices = np.where(clusters == cluster_id)[0]
        adj_submatrix = adjacency[cluster_indices, :][:, cluster_indices]
        degrees_submatrix = degrees[cluster_indices]
        result += np.sum(adj_submatrix) - (np.sum(degrees_submatrix)**2) / n_edges
    return result / n_edges


def conductance(adjacency, clusters):
    """Computes graph conductance as in Yang & Leskovec (2012).
    Args:
        adjacency: Input graph in terms of its sparse adjacency matrix.
        clusters: An (n,) int cluster vector.
    Returns:
        The average conductance value of the graph clusters.
    """
    inter = 0  # Number of inter-cluster edges.
    intra = 0  # Number of intra-cluster edges.
    cluster_indices = np.zeros(adjacency.shape[0], dtype=np.bool)
    for cluster_id in np.unique(clusters):
        cluster_indices[:] = 0
        cluster_indices[np.where(clusters == cluster_id)[0]] = 1
        adj_submatrix = adjacency[cluster_indices, :]
        inter += np.sum(adj_submatrix[:, cluster_indices])
        intra += np.sum(adj_submatrix[:, ~cluster_indices])
    return intra / (inter + intra)

  # -----------------------------------------------------------------------------

def normalize_graph(graph,
                    normalized = True,
                    add_self_loops = True):
    """Normalized the graph's adjacency matrix in the scipy sparse matrix format.
    Args:
        graph: A scipy sparse adjacency matrix of the input graph.
        normalized: If True, uses the normalized Laplacian formulation. Otherwise,
        use the unnormalized Laplacian construction.
        add_self_loops: If True, adds a one-diagonal corresponding to self-loops in
        the graph.
    Returns:
        A scipy sparse matrix containing the normalized version of the input graph.
    """
    if add_self_loops:
        graph = graph + sp.sparse.identity(graph.shape[0])
    degree = np.squeeze(np.asarray(graph.sum(axis=1)))
    if normalized:
        with np.errstate(divide='ignore'):
            inverse_sqrt_degree = 1. / np.sqrt(degree)
        inverse_sqrt_degree[inverse_sqrt_degree == np.inf] = 0
        inverse_sqrt_degree = sp.sparse.diags(inverse_sqrt_degree)
        return inverse_sqrt_degree @ graph @ inverse_sqrt_degree
    else:
        with np.errstate(divide='ignore'):
            inverse_degree = 1. / degree
        inverse_degree[inverse_degree == np.inf] = 0
        inverse_degree = sp.sparse.diags(inverse_degree)
        return inverse_degree @ graph

# -----------------------------------------------------------------------------

def load_npz(
    filename
):
    """Loads an attributed graph with sparse features from a specified Numpy file.
    Args:
        filename: A valid file name of a numpy file containing the input data.
    Returns:
        A tuple (graph, features, labels, label_indices) with the sparse adjacency
        matrix of a graph, sparse feature matrix, dense label array, and dense label
        index array (indices of nodes that have the labels in the label array).
    """
    with np.load(open(filename, 'rb'), allow_pickle=True) as loader:
        loader = dict(loader)
        adjacency = sp.sparse.csr_matrix(
            (loader['adj_data'], loader['adj_indices'], loader['adj_indptr']),
            shape=loader['adj_shape'])

        features = sp.sparse.csr_matrix(
            (loader['feature_data'], loader['feature_indices'],
            loader['feature_indptr']),
            shape=loader['feature_shape'])

        label_indices = loader['label_indices']
        labels = loader['labels']
    assert adjacency.shape[0] == features.shape[
        0], 'Adjacency and feature size must be equal!'
    assert labels.shape[0] == label_indices.shape[
        0], 'Labels and label_indices size must be equal!'
    return adjacency, features, labels, label_indices

# -----------------------------------------------------------------------------

def convert_scipy_sparse_to_sparse_tensor(
    matrix):
    """Converts a sparse matrix and converts it to Tensorflow SparseTensor.
    Args:
        matrix: A scipy sparse matrix.
    Returns:
        A ternsorflow sparse matrix (rank-2 tensor).
    """
    matrix = matrix.tocoo()
    return tf.sparse.SparseTensor(
        np.vstack([matrix.row, matrix.col]).T, matrix.data.astype(np.float32),
        matrix.shape)

# -----------------------------------------------------------------------------

def build_dmon(input_features,
               input_graph,
               input_adjacency,
               n_clusters,
               reg,
               dropout,
              ):
    """Builds a Deep Modularity Network (DMoN) model from the Keras inputs.
    Args:
        input_features: A dense [n, d] Keras input for the node features.
        input_graph: A sparse [n, n] Keras input for the normalized graph.
        input_adjacency: A sparse [n, n] Keras input for the graph adjacency.
    Returns:
        Built Keras DMoN model.
    """
    output = input_features
    for n_channels in [64]: # architecture
        output = GCN(n_channels)([output, input_graph])
    pool, pool_assignment = DMoN(n_clusters, collapse_regularization=reg, dropout_rate=dropout)([output, input_adjacency])
    return tf.keras.Model(
        inputs=[input_features, input_graph, input_adjacency],
        outputs=[pool, pool_assignment])

# -----------------------------------------------------------------------------


from SpatialCluster.utils.data_format import data_format, position_data_format
from SpatialCluster.utils.get_areas import get_areas
from scipy.stats import hypergeom
from scipy import spatial
import pandas as pd
import numpy as np

def KNN_Clustering(features_X, features_position, attribute, threshold, location, condition=">", k=5, K=30, alfa = 0.01, leafsize=10):
    features_X = data_format(features_X)
    features_position = position_data_format(features_position)
    features_X = pd.concat([features_position, features_X], axis=1)
    t = []
    f = []
    x_name = []
    x_centroid = []
    for name, df in features_X.groupby(location):
        x_name.append(name)
        t.append(df.shape[0])
        if(condition == ">"):
            f.append(df[df[attribute] > threshold].shape[0])
        elif(condition == "<"):
            f.append(df[df[attribute] < threshold].shape[0])
        elif(condition == ">="):
            f.append(df[df[attribute] >= threshold].shape[0])
        elif(condition == "<="):
            f.append(df[df[attribute] <= threshold].shape[0])
        elif(condition == "=="):
            f.append(df[df[attribute] == threshold].shape[0])
        x_centroid.append((df["lon"].mean(), df["lat"].mean()))
    tree = spatial.KDTree(data = x_centroid, leafsize = leafsize) # Revisar si hacerlo con centroides o con todos los puntos
    Nk = tree.query(x_centroid, k = k)[1]
    NK = tree.query(x_centroid, k = K)[1]

    Tk = []
    Fk = []
    for j in range(Nk.shape[0]):
        ti = 0
        fi = 0
        for i in Nk[j]:
            ti += t[i]
            fi += f[i]
        Tk.append(ti)
        Fk.append(fi)
    TK = []
    FK = []
    for j in range(NK.shape[0]):
        ti = 0
        fi = 0
        for i in NK[j]:
            ti += t[i]
            fi += f[i]
        TK.append(ti)
        FK.append(fi)

    p_hot = []
    p_cold = []
    for j in range(len(x_centroid)): 
        hpd = hypergeom(TK[j], FK[j], Tk[j])
        p = 0
        for z in range(Fk[j], Tk[j]):
            p += hpd.pmf(z)
        p_hot.append(p)
        p = 0
        for z in range(Fk[j]):
            p += hpd.pmf(z)
        p_cold.append(p)

    centroid_clusters = {}
    for i in range(len(x_centroid)):
        if(p_hot[i] < alfa):
            centroid_clusters[x_name[i]] = 1
        elif(p_cold[i] < alfa):
            centroid_clusters[x_name[i]] = -1
        else:
            centroid_clusters[x_name[i]] = 0
    
    clusters = np.zeros(features_X.shape[0])
    for i, row in features_X.iterrows():
        clusters[i] = centroid_clusters[row[location]]

    points = list(zip(features_X.lon, features_X.lat))
    areas_to_points = get_areas(clusters, points)
    return areas_to_points, clusters

    


import pandas as pd
import numpy as np

def data_format(data):
    try:
        data = pd.DataFrame(data)
        data = data.reset_index(drop=True)
        return data
    except:
        raise ValueError("Data not compatible with pandas DataFrame format")

def numpy_data_format(data):
    try:
        data = np.array(data)
        return data
    except:
        raise ValueError("Data not compatible with numpy Matrix format")

def position_data_format(data):
    try:
        data = pd.DataFrame(data, columns=["lon","lat"])
        data = data.reset_index(drop=True)
        return data
    except:
        raise ValueError("Data not compatible with pandas DataFrame format")



from SpatialCluster.constants import COLORS, palette
from matplotlib.colors import ListedColormap
import matplotlib.pyplot as plt
import matplotlib.colors as cl
import contextily as cx
from folium import Map
import numpy as np
import geopandas
import folium
import random

def plot_map(df, markersize = 30, figsize = (12,8)):
    if("geometry" not in df.columns):
        gdf = geopandas.GeoDataFrame(df, geometry=geopandas.points_from_xy(df['lon'], df['lat']))
    else:
        gdf = geopandas.GeoDataFrame(df, geometry='geometry', crs=4326)
    n_clusters = np.unique(gdf["clusters"]).shape[0]
    if(n_clusters > len(palette)):
        cmap = None
    else:
        cmap = ListedColormap(palette[:n_clusters])
    fig, ax = plt.subplots(1, figsize=figsize)
    
    gdf.plot(column="clusters", cmap=cmap, ax=ax, markersize=markersize)

    try:
        cx.add_basemap(ax, crs=gdf.to_crs(4326).crs.to_string())
    except:
        print(f'Basemap error')

    plt.show()


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

def plot_map_sample(areas_to_points, min_supp, max_samples_per_clusters, location = [-33.45, -70.65], radius = 10):
    COLORS_9_SET = [plt.cm.Set1(i) for i in range(10)]
    COLORS_9_SET = [cl.to_hex(c) for c in COLORS_9_SET]

    hmap = Map(location=location, control_scale=True, zoom_start=11, tiles = 'stamen toner')
    if len([x for x in areas_to_points.items() if len(x[1]) > min_supp]) <= 9:
        colors_to_use = COLORS_9_SET
    else:
        colors_to_use = COLORS
    sorted_points = {k: v for k, v in sorted(areas_to_points.items(), key=lambda item: len(item[1]), reverse=True)}
    
    for index, i in enumerate(sorted_points):
        if len(sorted_points[i]) <= min_supp:
            continue
        
        a_ = random.sample(sorted_points[i], min(len(sorted_points[i]), max_samples_per_clusters) )
        for point in a_:
            folium.Circle(location=[point[1], point[0]], popup = str(point),
                            color=colors_to_use[index], radius=radius).add_to(hmap)
            
    return hmap


from SpatialCluster.utils.data_format import data_format, numpy_data_format
from sklearn.metrics import adjusted_rand_score
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import numpy as np

def ARI(clustering_1, clustering_2):
    clustering_1 = numpy_data_format(clustering_1)
    clustering_2 = numpy_data_format(clustering_2)
    return adjusted_rand_score(clustering_1, clustering_2)

def ARI_matrix(clusterings, plot=True, figsize = (10,10), linewidth=1):
    clusterings = data_format(clusterings)
    ARS_array = []
    len_columns = len(clusterings.columns)
    for i, cluster_1 in enumerate(clusterings.columns): 
        for j in range(i+1, len_columns):
            cluster_2 = clusterings.columns[j]
            df_cluster_1 = clusterings[cluster_1]
            df_cluster_2 = clusterings[cluster_2]
            ARS = adjusted_rand_score(df_cluster_1, df_cluster_2)
            ARS_array.append([cluster_1, cluster_2, ARS])
            ARS_df = pd.DataFrame(ARS_array)
            ARS_df_pivot = ARS_df.pivot(index=0, columns=1, values=2,).fillna(0)
    if(plot):
        plt.figure(figsize = figsize)
        mask = np.zeros_like(ARS_df_pivot)
        mask[np.triu_indices_from(mask)] = True
        g = sns.heatmap(
            ARS_df_pivot, 
            cmap='OrRd',
            linewidth=linewidth,
            mask=mask,
            annot=True, fmt="0.2f")
        g.set_title('Adjusted Rand Score')
        g.set_xticklabels(g.get_xticklabels(), rotation=45, horizontalalignment='right')
        plt.show()
    return ARS_df_pivot


import scipy as sp
import numpy as np
import array

class IncrementalCOOMatrix(object):

    def __init__(self, shape, dtype):
        if dtype is np.int32:
            type_flag = 'i'
        elif dtype is np.int64:
            type_flag = 'l'
        elif dtype is np.float32:
            type_flag = 'f'
        elif dtype is np.float64:
            type_flag = 'd'
        else:
            raise Exception('Dtype not supported for matrix (int32, int64, float32, float64). ')

        self.dtype = dtype
        self.shape = shape

        self.rows = array.array('i')
        self.cols = array.array('i')
        self.data = array.array(type_flag)

    def append(self, i, j, v):
        m, n = self.shape
        if (i >= m or j >= n):
            raise Exception('Index out of bounds')

        self.rows.append(i)
        self.cols.append(j)
        self.data.append(v)

    def tocoo(self):
        rows = np.frombuffer(self.rows, dtype=np.int32)
        cols = np.frombuffer(self.cols, dtype=np.int32)
        data = np.frombuffer(self.data, dtype=self.dtype)

        return sp.sparse.coo_matrix((data, (rows, cols)),
                             shape=self.shape)

    def __len__(self):

        return len(self.data)

# -----------------------------------------------------------------------------